from .sig_sys import kill, cpu_percent, mem_percent, mem_info
from .sig_queue import list_queue, queue_size, queue_get_switch, queue_put_switch
from .sig_event import node_connect, heartbeat, health_check, health_check_for_all

# 信号组
signal_group = {
    "sys": [kill, cpu_percent, mem_percent, mem_info],
    "queue": [list_queue, queue_size, queue_get_switch, queue_put_switch],
    "event": [node_connect, heartbeat, health_check, health_check_for_all]
}

all_signal = []
for k in signal_group.keys():
    all_signal.extend(signal_group.get(k, []))



import time
from multiprocessing.managers import RemoteError
from loguru import logger
import psutil
import os
from .exceptions import SignalException
from .base import fn_get, fn_qsize, fn_put, fn_clear, EHBaseManager
from .master import signal
from .util.net import get_mac_addr


class Node(EHBaseManager):

    def __init__(self, name, address=None, authkey=None, serializer='pickle',
                 ctx=None, **kwargs):
        super().__init__(name, address=address, authkey=authkey, serializer=serializer, ctx=ctx, **kwargs)
        self.mac = get_mac_addr()
        self.fps = 0

        self._send_heartbeat_interval = 2
        self._send_heartbeat_at = 0
        self._process = psutil.Process(os.getpid())

        # 注册的队列名列表
        self.register_queue_names = []
        self.connect()
        # 发送链接成功的信号
        self.send_master_signal(signal.node_connect, name, self.mac)

    def register_queue(self, queue_name):
        """
        注册队列
        注册后的队列可以被写入和读取
        :param queue_name:
        :return:
        """
        self._register_queue(queue_name)
        self.register(fn_get(queue_name))
        self.register(fn_put(queue_name))
        self.register(fn_clear(queue_name))
        self.register(fn_qsize(queue_name))
        self.register_queue_names.append(queue_name)

    def put(self, data, queue_name):
        """
        将数据发送到指定队列
        :param data:
        :param queue_name:
        :return:
        """
        getattr(self, fn_put(queue_name))(data)

    def get(self, queue_name) -> dict:
        """
        从指定队列获取数据
        :param queue_name:
        :return:
        """
        item = getattr(self, fn_get(queue_name))()
        if item:
            return item._getvalue()

    def clear(self, queue_name):
        """
        清空指定队列
        :param queue_name:
        :return:
        """
        getattr(self, fn_clear(queue_name))()

    def qsize(self, queue_name) -> int:
        """
        获取队列的大小
        :param queue_name:
        :return:
        """
        size = getattr(self, fn_qsize(queue_name))()
        return size

    def send_master_signal(self, sig, *args, **kwargs):
        """
        向master节点发送信号
        :param sig:
        :param args:
        :param kwargs:
        :return:
        """
        if hasattr(sig, "__call__"):
            sig = sig.__name__
        _r = self._send_master_signal(sig, *args, **kwargs)
        data = _r._getvalue()
        if data['is_ok']:
            result = data['result']
            return result
        else:
            raise SignalException(str(data['msg']))

    def run(self):
        logger.info("register queue:")
        for name in self.register_queue_names:
            logger.info(f"- {name}")
        # before run生命周期回调
        self.before_run()

        while True:
            try:
                _tic = time.time()
                self._send_heartbeat()
                self.on_all_queue_process()
                # 更新fps
                self.fps = 1 / (time.time() - _tic)
            except RemoteError as e:
                """
                队列数据被消费完时会触发此异常
                """
                time.sleep(0.1)

    def _send_heartbeat(self):
        """
        发送心跳信号
        :return:
        """
        now = time.time()
        if now - self._send_heartbeat_at > self._send_heartbeat_interval:
            self.send_master_signal(signal.heartbeat, self.name, self.performance_info())
            self._send_heartbeat_at = now

    def before_run(self):
        """
        run执行前的回调 适合执行一些前置初始化
        :return:
        """
        pass

    def on_all_queue_process(self):
        """
        所有队列需要被处理的时的回调
        :return:
        """
        # 单个队列的处理回调(常用)
        for name in self.register_queue_names:
            self.on_queue_process(name)
            logger.debug(f"qsize: {self.qsize(name)}")

    def on_queue_process(self, queue_name):
        """
        队列被处理的回调入口
        :param queue_name:
        :return:
        """
        pass

    def performance_info(self) -> dict:
        """
        节点实时性能数据
        :return:
        """
        return {
            "base": {
                "fps": self.fps
            },
            "cpu": {
                "percent": self._process.cpu_percent()
            },
            "mem": {
                "percent": self._process.memory_percent(),
                "info": self._process.memory_info()
            }
        }

    def start_print(self):
        super(Node, self).start_print()
        print(f"Node ==> Master {self.address}")
        self.print_line()



"""
事件相关信号
"""
from ..node_info import NodeInfo
from loguru import logger


def node_connect(master, name, client_info):
    """
    节点连接到master
    :param master:
    :param name:
    :param client_info:
    :return:
    """
    _connect_pre_check(master, name)
    master.nodes[name] = NodeInfo(name, client_info)
    logger.info(f">>> Node [{name}]({client_info}) has successfully connected to the master!")


def heartbeat(master, name, performance_info):
    """
    心跳事件
    :param master:
    :param name:
    :param performance_info:
    :return:
    """
    # 刷新节点的时间记录
    logger.debug(f"recv {name} heartbeat")
    ni = master.nodes.get(name)
    ni.recv(performance_info)


def health_check(master, name):
    """
    检查指定节点的健康情况
    :param master:
    :param name:
    :return:
    """
    ni = master.nodes.get(name)
    return ni.is_health()


def health_check_for_all(master):
    """
    所有节点的健康检查
    :param master:
    :return:
    """
    r = {}
    for node in master.nodes.keys():
        r[node] = master.nodes[node].is_health()
    return r


def _connect_pre_check(master, name):
    # if name in master.nodes.keys():
    #     raise ConnectionError(f"{name} has already connected.")
    pass



"""
队列操作相关信号
"""
from loguru import logger
from ...base import EQueue


def list_queue(master):
    """
    list all queue
    :param master:
    :return:
    """
    return master.queue_dict.keys()


def queue_size(master, queue_name):
    """
    get queue size
    :param master:
    :param queue_name:
    :return:
    """
    if queue_name in master.queue_dict.keys():
        q: EQueue = master.queue_dict.get(queue_name)
        return q.qsize()
    else:
        logger.error(f"{queue_name} is not exist in master")
        return 0


def queue_put_switch(master, queue_name, cmd="enable"):
    """
    control the enable and disable of the put action of the queue
    :param master:
    :param queue_name:
    :param cmd:
    :return:
    """
    if queue_name in master.queue_dict.keys():
        q: EQueue = master.queue_dict.get(queue_name)
        if cmd == "enable":
            q.enable_put = True
        elif cmd == "disable":
            q.enable_put = False
        else:
            logger.error(f"put switch cmd error, cmd should be 'enable' or 'disable', but your value is {cmd}")
    else:
        logger.error(f"{queue_name} is not exist in master")


def queue_get_switch(master, queue_name, cmd="enable"):
    """
    control the enable and disable of the get action of the queue
    :param master:
    :param queue_name:
    :param cmd:
    :return:
    """
    if queue_name in master.queue_dict.keys():
        q: EQueue = master.queue_dict.get(queue_name)
        if cmd == "enable":
            q.enable_get = True
        elif cmd == "disable":
            q.enable_get = False
        else:
            logger.error(f"get switch cmd error, cmd should be 'enable' or 'disable', but your value is {cmd}")
    else:
        logger.error(f"{queue_name} is not exist in master")



from multiprocessing.managers import BaseManager

from queue import Queue
import time
import uuid
import os
import logging
import sys
from loguru import logger
from .util.loghandle import InterceptHandler

fn_put = lambda name: f"_{name}_put"
fn_get = lambda name: f"_{name}_get"
fn_clear = lambda name: f"_{name}_clear"
fn_qsize = lambda name: f"_{name}_qsize"


def message(data):
    return {"data": data, "at": time.time(), "id": str(uuid.uuid4())}


_logo = \
    """
      _____    _            _   _       _     
     | ____|__| | __ _  ___| | | |_   _| |__  
     |  _| / _` |/ _` |/ _ \ |_| | | | | '_ \ 
     | |__| (_| | (_| |  __/  _  | |_| | |_) |
     |_____\__,_|\__, |\___|_| |_|\__,_|_.__/ 
                 |___/                        
    """


class EQueue(Queue):

    def __init__(self):
        super().__init__()
        self.enable_put = True
        self.enable_get = True

    def put(self, item, block=True, timeout=None):
        if self.enable_put:
            super(EQueue, self).put(item, block=block, timeout=timeout)

    def get(self, block=True, timeout=None):
        if self.enable_get:
            return super(EQueue, self).get(block=block, timeout=timeout)


class EHBaseManager(BaseManager):
    def __init__(self, name, address=None, authkey=None, serializer='pickle',
                 ctx=None, **kwargs):
        super().__init__(address=address, authkey=authkey, serializer=serializer, ctx=ctx)
        self.name = name
        self.start_print()
        self.register("_register_queue", callable=self._register_queue)
        self.register("_send_master_signal", callable=self._send_master_signal)
        self._init_log(kwargs.get("log_level", "INFO"), log_dir=kwargs.get("log_dir", "logs"))

    def _init_log(self, log_level, log_dir):
        logging.basicConfig(handlers=[InterceptHandler()], level=log_level)
        logger.configure(handlers=[{"sink": sys.stderr, "level": log_level}])

        t_name = type(self).__name__
        logger.add(
            os.path.join(log_dir, f"{t_name}-{self.name}-" + "{time}.log"),
            level=log_level,
            rotation="00:00",
            retention="10 days",
            backtrace=True,
            diagnose=True,
            enqueue=True,
        )

    def _register_queue(self, queue_name):
        """
        注册队列
        :param queue_name:
        :return:
        """

    def _send_master_signal(self, signal, *args, **kwargs):
        """
        发送信号给master
        :param signal:
        :param args:
        :param kwargs:
        :return:
        """

    def start_print(self):
        from . import __version__
        print()
        self.print_line()
        print(_logo)
        print(f"Version: {__version__}")

    @classmethod
    def print_line(cls):
        print("---" * 14)



"""
系统控制相关信号
"""
import os
import psutil
from ...util.kill import kill as _kill


def kill(master):
    """
    kill master
    :param master:
    :return:
    """
    _kill(os.getpid())


def cpu_percent(master):
    """
    get cpu percent info
    :return:
    """
    pid = os.getpid()
    p = psutil.Process(pid)
    return p.cpu_percent()


def mem_percent(master):
    """
    get memory percent info
    :return:
    """
    pid = os.getpid()
    p = psutil.Process(pid)
    return p.memory_percent()


def mem_info(master):
    """
    get memory info
    :return:
    """
    pid = os.getpid()
    p = psutil.Process(pid)
    return p.memory_info()



import logging
from types import FrameType
from typing import cast

from loguru import logger


class InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:  # pragma: no cover
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = str(record.levelno)

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:  # noqa: WPS609
            frame = cast(FrameType, frame.f_back)
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage(),
        )



import time


class NodeInfo:
    def __init__(self, name, client_info):
        self.name = name
        self.client_info = client_info
        self._last_recv_at = time.time()
        self.performance_info = {}

    def recv(self, performance_info):
        self._last_recv_at = time.time()
        self.performance_info = performance_info

    def is_health(self, timeout_duration=10):
        """
        是否正常(默认10秒没有收到心跳 则为不正常)
        :param timeout_duration:
        :return:
        """
        return time.time() - self._last_recv_at < timeout_duration



import os


def kill(pid):
    # 本函数用于中止传入pid所对应的进程
    if os.name == 'nt':
        # Windows系统
        cmd = 'taskkill /pid ' + str(pid) + ' /f'
        try:
            os.system(cmd)
        except Exception as e:
            print(e)
    elif os.name == 'posix':
        # Linux系统
        cmd = 'kill ' + str(pid)
        try:
            os.system(cmd)
        except Exception as e:
            print(e)
    else:
        print('Undefined os.name')



#!/usr/bin/env python3
# _*_ coding: utf-8 _*_
__author__ = "DannyLee1991"

from setuptools import setup, find_packages
from edgehub import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='edgehub',
      version=__version__,
      description='A flexible and efficient cross process distributed edge computing engine.',
      author='DannyLee1991',
      author_email='747554505@qq.com',
      url='https://github.com/EdgeGalaxy/edgehub',
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type="text/markdown",
      license="GPLv3",
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent"],

      python_requires='>=3.6.9',
      install_requires=[
          "loguru",
          "psutil>=5.9.0",
      ]
      )



from ..base import fn_put, fn_get, fn_qsize, fn_clear, EHBaseManager, message, EQueue
from .signal import all_signal
from loguru import logger


class Master(EHBaseManager):

    def __init__(self, name, address=None, authkey=None, serializer='pickle',
                 ctx=None, **kwargs):
        super().__init__(name, address=address, authkey=authkey, serializer=serializer, ctx=ctx, **kwargs)
        # key node_name => val node_info obj
        self.nodes = {}
        self.queue_dict = {}

    def _send_master_signal(self, signal, *args, **kwargs):
        """
        信号执行入口
        :param signal:
        :param args:
        :param kwargs:
        :return:
        """
        logger.debug(f"receive signal: {signal} args: {args} kwargs: {kwargs}")
        is_ok = False
        result = ""
        for sig in all_signal:
            if signal == sig.__name__:
                try:
                    result = sig(self, *args, **kwargs)
                    msg = "success"
                    is_ok = True
                except Exception as e:
                    is_ok = False
                    result = ""
                    msg = str(e)
                break
        else:
            msg = f"unknown signal {signal}"
        return {"is_ok": is_ok, "result": result, "msg": msg}

    def _register_queue(self, queue_name):
        """
        注册队列
        :param queue_name:
        :return:
        """
        logger.info(f"register queue: {queue_name}")

        def put_func(data, block=True, timeout=None):
            q: EQueue = self.queue_dict.get(queue_name)
            q.put(message(data), block=block, timeout=timeout)

        def get_func(block=False, timeout=None) -> dict:
            q: EQueue = self.queue_dict.get(queue_name)
            item = q.get(block=block, timeout=timeout)
            data = item.get("data")
            return data

        def clear_func():
            q: EQueue = self.queue_dict.get(queue_name)
            while not q.empty():
                q.get_nowait()

        def qsize_func() -> int:
            q: EQueue = self.queue_dict.get(queue_name)
            return q.qsize()

        # 添加队列
        if queue_name not in self.queue_dict.keys():
            self.queue_dict[queue_name] = EQueue()
            self.register(fn_put(queue_name), callable=put_func)
            self.register(fn_get(queue_name), callable=get_func)
            self.register(fn_clear(queue_name), callable=clear_func)
            self.register(fn_qsize(queue_name), callable=qsize_func)

    def start_print(self):
        super(Master, self).start_print()
        print(f"Master {self.address}")
        self.print_line()
